" autoload/selx.vim
"
" SelX - Richard Bentley-Green, 27/10/2020
"
" Additional highlighting functions - These work similarly to the normal '*'
" function but multiple highlights (in different colours) are supported.
" These highlights operate on a per-tab basis
"
if exists('g:Selx_loaded_autoload')
  finish
endif
let g:Selx_loaded_autoload = 1

" ------------------------------------------------------------------------------
" A table of highlight colours that may be set. This may be any length > 0
if exists('g:Selxlist')
  let s:selXlist = g:Selxlist
else
  let s:selXlist = ['SelX1', 'SelX2', 'SelX3', 'SelX4', 'SelX5', 'SelX6', 'SelX7', 'SelX8', 'SelX9']
endif

" Timer used for controlling display of window status line info
let s:showselx = 0

" Set timer for displaying SelX selection options
function selx#SetShowSelX()
  let s:showselx = 3
endfunction

" This is a non-standard colour
highlight default link StatusMsg Normal

" Print a 'status' message to the command line
function s:PrintStatusMsg(msg)
  echohl StatusMsg | echo a:msg | echohl None
endfunction

" Print a 'warning' message to the command line
function s:PrintWarningMsg(msg)
  echohl WarningMsg | echo a:msg | echohl None
endfunction

" Generate and return text that may be incorporated into window status line to show
" SelX highlights
"
" bgcol - Main colour of window status line
"
" Returns either an empty string (if highlight display is not appropriate) or
" a string containing the highlight info to incorporate into the window status
" line
"
function! selx#Statusline(bgcol)
  let l:selx = ''
  if s:showselx
    let s:showselx -= 1
    let l:dispx = exists('t:selXstr') && exists('w:selXlist')
    let l:nextIdx = (exists('t:nextSelX')) ? t:nextSelX : 0
    let l:selx = '['
    let l:idx = 0
    while (l:idx < len(s:selXlist))
      let l:next = (l:idx == l:nextIdx) ? '^' : ' '
      let l:selx = l:selx . '%#' . s:selXlist[l:idx] . '#' . l:next . (l:dispx && (w:selXlist[l:idx] > 0) ? (l:idx + 1) . ' ' : '  ')
      let l:idx += 1
    endwhile
    let l:selx = l:selx . a:bgcol . '] '
  endif
  return l:selx
endfunction

" Print details of all active SelX highlights
"
" all - 1 = Print details of all highlights (if at
"           leat one is active)
"       0 = Only print details of active highlights
"
function selx#ActiveList(all)
  let l:active = 0
  if exists('t:selXstr') && exists('w:selXlist')
    let l:nextIdx = (exists('t:nextSelX')) ? t:nextSelX : 0
    let l:idx = 0
    while (l:idx < len(s:selXlist))
      if a:all || (w:selXlist[l:idx] > 0)
        let l:next = (l:idx == l:nextIdx) ? '^' : ' '
        exe 'echohl ' . s:selXlist[l:idx]
        echo l:next . ((w:selXlist[l:idx] > 0) ? (l:idx + 1) . ' ' : '  ')
        echohl None
        echon ' ' . t:selXstr[l:idx]
        let l:active = 1
      endif
      let l:idx += 1
    endwhile
  endif

  if !l:active
    call s:PrintStatusMsg('SelX: No active selections')
  endif
endfunction

" Set/refresh highlight for all windows
"
" sel - Text to match
" idx - The colour to use (0 to N-1).
"
function s:SelX(sel, idx)
  " Save view before iterating through the windows
  let l:focuswin = winnr()
  let l:winview = winsaveview()

  " Set the highlight (at a priority below the normal search highlight)
  for l:winx in range(1, winnr('$'))
    exe 'keepjumps noautocmd ' . l:winx . 'wincmd w'
    if !exists('w:selXlist')
      let w:selXlist = repeat([0], len(s:selXlist))
    else
      if w:selXlist[a:idx] > 0
        " Selection colour already in use - clear selection.
        " In some complex cases, after just splitting a window, and
        " being called from SelXRefresh(), the match index may not
        " exist, hence the 'silent'
        silent! call matchdelete(w:selXlist[a:idx])
      endif
    endif

    " Activate new highlight selection
    let w:selXlist[a:idx] = matchadd(s:selXlist[a:idx], a:sel, -1)
  endfor

  " Restore view
  exe 'keepjumps noautocmd ' . l:focuswin . 'wincmd w'
  call winrestview(l:winview)
endfunction

" Locate the next highlight colour to use for auto-placement,
" and adjust t:nextSelX accordingly
function s:SelXnext()
  let l:stop = 0
  let l:next = t:nextSelX
  while (!l:stop) && (t:selXstr[l:next] != '')
    let l:next += 1
    if l:next >= len(s:selXlist)
      let l:next = 0
    endif
    let l:stop = (l:next == t:nextSelX) || (t:selXnum == len(s:selXlist))
  endwhile
  let t:nextSelX = l:next
endfunction

" Set highlight in next available colour
"
" sel - Selected text. If '', then current (whole) word is assumed
" idx - If non-zero then this indicates the colour to use (1 to N).
"       If zero then the next colour is selected automatically
"
" Returns the index that was used (zero on an error)
"
function selx#SelXAdd(sel, idx)
  let l:used=0
  let l:text = a:sel
  if l:text == ''
    " No text specified - assume current (whole!) word
    let l:text = '\<'.expand("<cword>").'\>'
    if l:text == '\<\>'
      let l:text = ''
    endif
  endif

  if l:text != ''
    if !exists('t:selXstr')
      let t:selXstr = repeat([''], len(s:selXlist))
      let t:nextSelX = 0
      let t:selXnum = 0
    endif

    " Determine the next selection colour to use
    if (a:idx > 0) && (a:idx <= len(s:selXlist))
      let t:nextSelX = a:idx - 1
    endif

    " Remember the selection
    if t:selXstr[t:nextSelX] == ''
      let t:selXnum += 1
    endif
    let t:selXstr[t:nextSelX] = l:text

    " Set the highlight
    let l:used = t:nextSelX + 1
    call s:SelX(l:text, t:nextSelX)

    " Step on to next highlight for next time
    call s:SelXnext()
  endif

  return l:used
endfunction

" Refresh all selection highlights in current window - required when a new window is added
function selx#SelXRefresh()
  if exists('t:selXstr') && t:selXnum
    if !exists('w:selXlist')
      let w:selXlist = repeat([0], len(s:selXlist))
    endif

    let l:idx = 0
    while (l:idx < len(t:selXstr))
      if t:selXstr[l:idx] != ''
        if w:selXlist[l:idx] > 0
          " Selection colour already in use - clear selection
          silent! call matchdelete(w:selXlist[l:idx])
        endif

        " Activate new highlight selection
        let w:selXlist[l:idx] = matchadd(s:selXlist[l:idx], t:selXstr[l:idx], -1)
      endif
      let l:idx += 1
    endwhile
  endif
endfunction

" Clear all, or a specified selection highlight
"
" idx - If zero then all selections are cleared. Otherwise this is
"       the index (1 to N) of the selection to clear
"
function selx#SelXClr(idx)
  if exists('t:selXstr') && t:selXnum && (a:idx >= 0) && (a:idx <= len(t:selXstr))
    " Save view before iterating through the windows
    let l:focuswin = winnr()
    let l:winview = winsaveview()

    if (!a:idx)
      " Clear all selections
      for l:winx in range(1, winnr('$'))
        exe 'keepjumps noautocmd ' . l:winx . 'wincmd w'
        if exists('w:selXlist')
          let l:idx = 0
          while (l:idx < len(w:selXlist))
            if (w:selXlist[l:idx] > 0)
              call matchdelete(w:selXlist[l:idx])
            endif
            let l:idx += 1
          endwhile

          unlet w:selXlist
        endif
      endfor
      let t:selXnum = 0
    else
      " Request to clear specific selection
      let l:idx = a:idx - 1
      if t:selXstr[l:idx] != ''
        " Selection is set - clear it
        for l:winx in range(1, winnr('$'))
          exe 'keepjumps noautocmd ' . l:winx . 'wincmd w'
          if exists('w:selXlist')
            if (w:selXlist[l:idx] > 0)
              call matchdelete(w:selXlist[l:idx])
              let w:selXlist[l:idx] = 0
              if t:selXnum == 1
                unlet w:selXlist
              endif
            endif
          endif
        endfor
        let t:selXstr[l:idx] = ''
        let t:selXnum -= 1

        " Clearing of this highlight might mean we have a spare colour to use
        " where we didn't previously
        if (t:selXstr[t:nextSelX] != '')
          let t:nextSelX = l:idx
        endif
      endif
    endif

    if (!t:selXnum)
      " Deleting these control variables makes further operations more efficient
      unlet t:selXstr
      unlet t:nextSelX
      unlet t:selXnum
    endif

    " Restore view
    exe 'keepjumps noautocmd ' . l:focuswin . 'wincmd w'
    call winrestview(l:winview)
  endif
endfunction

" Return the text associated with the specified highlight colour
"
" idx - The colour to use (1 to N)
"
" Returns associated string. An empty string is returned on an error or if
" the specified colour is not assigned
"
function s:SelXGetText(idx)
  let l:text = ''
  if exists('t:selXstr') && (a:idx > 0) && (a:idx <= len(s:selXlist))
    let l:idx = a:idx - 1
    let l:text = t:selXstr[l:idx]
  endif
  return l:text
endfunction

" Same thing but using a specific colour (ie, '<leader>&N', '<leader>:N')

" These commands operate in two parts. This allows the highlight map to be
" displayed before the command entry completes (ie, before the 'N' value is
" entered, which makes the commands easier to use). It also gets around the
" problem whereby the getchar() function (used to implement the 'display map
" before completing the command' feature) moves the cursor to the command
" entry line (which is otherwise annoying)
"
" To do this, the function SelXCmdInit() is called from the keyboard mapping.
" This displays the highlight map and then injects a new key sequence into
" the keyboard buffer that in-turn causes the second function to be called;
" SelXCmdExecute(). This second function reads the 'N' value via getchar() and
" then invokes whatever it needs to (as specified by the mapping) to complete
" the operation
"
" To support this, each command needs three keyboard mapping; the first is the
" one the user actually invokes, and this calls the SelXCmdInit() function.
" The second mapping is invoked by SelXCmdInit() via the keyboard sequence it
" injects. The third mapping is never actually used in normal operation; its
" sole purpose is to make the second mapping deliberately ambiguous (the
" ambiguity is resolved only when the 'N' key is pressed). In this way, the
" invocation of SelXCmdExecute() is delayed until there is guaranteed to be a
" key for getchar() to read, thus preventing it from blocking and thus
" stopping the cursor temporarily moving to the command line. Although not its
" primary purpose, the use of the third mapping also allows a clean break-out
" of the mapping (via <esc> or (practically) any other non-mapped character)
"
function selx#SelXCmdInit(cmdKey)
  " Display highlight map in status line
  call selx#SetShowSelX()
  redrawstatus
  call feedkeys(a:cmdKey)
endfunction

let selx#SelXCmdIndex = 0
function selx#SelXCmdExecute(expr, expr2)
  " Get which highlight to operate on (1 - N)
  let selx#SelXCmdIndex = str2nr(nr2char(getchar()))

  " Invoke operation
  exe a:expr
  if a:expr2 != ''
    exe a:expr2
  endif
endfunction

" Set a highlight
" A single string is supplied (which may contain spaces). The next
" available highlight colour shall be used
function selx#SelXAddCmd(str)
  if (a:str != '')
    let l:used = selx#SelXAdd(a:str, 0)
    call s:PrintStatusMsg('SelX: Index '.l:used.' set for "'.a:str.'"')
    call selx#SetShowSelX()
  else
    call s:PrintWarningMsg("usage: SelX <text>")
  endif
endfunction

" Set a highlight
" A single string is supplied in the format "N xxx" (where N is the highlight
" index and xxx is the string to match (which may contain spaces)
function selx#SelXNAddCmd(str)
  let l:err=1
  if (a:str[1:1] == ' ') && (a:str[2:2] != ' ')
    " str2nr(N) will return zero if N is not numeric
    let l:idx = str2nr(a:str[0:0])
    if (l:idx > 0) && (l:idx <= len(s:selXlist))
      let l:str = a:str[2:]
      let l:used = selx#SelXAdd(l:str, l:idx)
      call s:PrintStatusMsg('SelX: Index '.l:used.' set for "'.l:str.'"')
      call selx#SetShowSelX()
      let l:err=0
    endif
  endif
  if (l:err)
    call s:PrintWarningMsg("usage: SelX <N> <text>")
  endif
endfunction

" ----------------------------------
" Save SelX highlight configuration to a file
"
" If there are no SelX highlights enabled then nothing shall be written
"
" filename - The file to write to. This should have a .vim' extension because
"            it's a vim script that can be 'sourced'
" append - If true then the file shall be appended. Otherwise it shall be
"          overwritten
"
function selx#SelXSave(filename, append)
  " Save view before iterating through the tabs
  let l:winview = winsaveview()
  let l:focustab=tabpagenr()

  let l:haveset = 0

  tabfirst
  let l:tabnum = 1

  " Iterate through tabs
  while (l:tabnum <= tabpagenr('$'))
    if exists('t:selXstr') && t:selXnum
      " Tab has some SelX values set
      if !l:haveset
        exe 'redir! '.((a:append == 0) ? '>' : '>>').a:filename
        silent exe "echo '".'" Start of SelX configuration'."'"
        silent exe "echo 'if exists(''g:Selx_loaded'')'"
        silent exe "echo '  let s:winview = winsaveview()'"
        silent exe "echo '  let s:focustab=tabpagenr()'"
        let l:haveset = 1
      endif
      let l:idx = 1
      while (l:idx <= len(s:selXlist))
        let l:sel = s:SelXGetText(l:idx)
        if l:sel != ''
          " SelX valus is set - save it
          silent exe "echo '  call selx#SelXRestoreFromFile('.l:tabnum.\", '\".l:sel.\"', \".l:idx.')'"
        endif
        let l:idx = l:idx + 1
      endwhile

      " Remember 'next highlight' index
      silent "echo '  call selx#SelXRestoreNextFromFile('.l:tabnum.', '.t:nextSelX.')'"
    endif

    " Next tab
    normal! gt
    let l:tabnum = l:tabnum + 1
  endwhile

   if l:haveset
    silent exe "echo '  exe \"tabnext \".s:focustab'"
    silent exe "echo '  call winrestview(s:winview)'"
    silent exe "echo 'endif'"
    silent exe "echo '".'" End of SelX configuration'."'"
    silent exe "echo ''"
    silent exe "echo ''"
    redir END
  endif

  " Restore view
  exe 'tabnext '.l:focustab
  call winrestview(l:winview)
endfunction

" Called by the vim script created by the SelXSave() function
"
" NOTE: This does not save the current state before (possibly) moving to another
" tab. The state is expected to be saved/restored as required in the script
" that calls this function
"
" tabnum - The tab number to set the highlight in
" sel - The matched text
" idx - The index of the highlight to set
"
function selx#SelXRestoreFromFile(tabnum, sel, idx)
  exe 'tabnext '.a:tabnum
  call selx#SelXAdd(a:sel, a:idx)
endfunction

" Called by the vim script created by the SelXSave() function
"
" NOTE: This does not save the current state before (possibly) moving to another
" tab. The state is expected to be saved/restored as required in the script
" that calls this function
"
" tabnum - The tab number to set the highlight in
" next - The index of the next highlight to use
"
function selx#SelXRestoreNextFromFile(tabnum, next)
  if a:next < len(s:selXlist)
    exe 'tabnext '.a:tabnum
    let t:nextSelX = a:next
  endif
endfunction

" ----------------------------------
" SEARCH

" Modify search pattern so that it works
function selx#VSetSearch(cmdtype, v)
  " Save 's' and yank highlighted text to it
  let l:temp = @s
  norm! gv"sy
  let @/ = ((a:v) ? '\V' : '') . substitute(escape(@s, a:cmdtype.'\'), '\n', '\\n', 'g')
  let @s = l:temp
endfunction

" ----------------------------------
" Search for (jump to) previous or next instance of specified pattern without
" changing the highlighting
" text - Selected text. If '', then current (whole) word is assumed
" fwd - 1 = search forwards, else search backwards
" igc - -1 = Use default 'ignorecase' setting
"       0 - DO not ignore case
"       1 - Ignore case
"
function selx#Search(text, fwd, igc)
  let l:text = a:text
  if l:text == ''
    " No text specified - assume current (whole!) word
    let l:text = '\<'.expand("<cword>").'\>'
  endif

  " Search
  let l:search = ((a:igc) ? '\c' : '\C')
  let l:search = l:search . l:text
  call search(l:search, (a:fwd) ? '' : 'b')
endfunction

" ----------------------------------
" Search for additional highlighted text
"
" Note: This works in the same way as the functions above (ref. SelXCmdInit())

" Move to next/previous instance of specified highlight
" fwd - 1 = search forwards, else search backwards
let s:SelXGoFwd = 1

function selx#SelXGoToInit(fwd)
  " Display highlight map in status line
  let s:SelXGoFwd = a:fwd
  call selx#SetShowSelX()
  redrawstatus
  call feedkeys("+\<M-/>")
endfunction

let s:SelXGoLastIndex = 0

function selx#SelXGoToNext(readIdx, reverse)
  " Get which highlight to search for (1 - N)
  let l:idx = (a:readIdx) ? str2nr(nr2char(getchar())) : s:SelXGoLastIndex
  let l:text = s:SelXGetText(l:idx)
  if l:text != ''
    let s:SelXGoLastIndex = l:idx
    call selx#Search(l:text, (a:reverse) ? 1 - s:SelXGoFwd : s:SelXGoFwd, 0)
  endif
endfunction

" ----------------------------------
" Toggle selection between 'whole word' and 'arbitrary'
" text - Text to adjust
"
" Returns modified text
"
function selx#ToggleSel(text)
  let l:text = a:text
  let l:len = strlen(a:text)

  " Note: 'len > 4' to filter-out empty 'whole word' - it makes no sense to toggle this
  if (l:len > 4) && (strpart(a:text, 0, 2) == '\<') && (strpart(a:text, l:len - 2, 2) == '\>')
    " Whole word - convert to selection
    let l:text = '\V'.strpart(a:text, 2, l:len - 4)
    call s:PrintStatusMsg('Selection: arbitrary')
  elseif (l:len > 2) && (strpart(a:text, 0, 2) == '\V')
    " Selection - convert to whole word
    let l:text = '\<'.strpart(a:text, 2).'\>'
    call s:PrintStatusMsg('Selection: whole word')
  endif
  " else empty selection or cannot determine - leave as-is
  return l:text
endfunction

" ----------------------------------
" Toggle selection between 'whole word' and 'arbitrary' for specified highlight
function selx#SelXToggleSelInit()
  " Display highlight map in status line
  call selx#SetShowSelX()
  redrawstatus
  call feedkeys("+\<M-(>")
endfunction

function selx#SelXToggleSelNext()
  " Get which highlight to toggle (1 - N)
  let l:idx = str2nr(nr2char(getchar()))
  let l:text = s:SelXGetText(l:idx)
  if l:text != ''
    let l:idx -= 1
    let t:selXstr[l:idx] = selx#ToggleSel(l:text)
    " It should be possible to call s:SelX() here but the refresh doesn't work
    " Don't know why
    call selx#SelXRefresh()
  endif
endfunction

" ----------------------------------
" WINDOW CONTROL

" Open a new tab with the current window full-screen within it. Also
" copy across any active 'SelX' highlights
"
function selx#OpenInNewTab()
  if exists('t:selXstr') && t:selXnum
    let l:selXstr = t:selXstr
    let l:nextSelX = t:nextSelX
    let l:selXnum = t:selXnum
    tab split

    " Need to copy the array of strings longhand to avoid inter-tab linkage
    let t:selXstr = repeat([''], len(s:selXlist))
    let l:idx = 0
    while (l:idx < len(l:selXstr))
      let t:selXstr[l:idx] = l:selXstr[l:idx]
      let l:idx = l:idx + 1
    endwhile

    let t:nextSelX = l:nextSelX
    let t:selXnum = l:selXnum
    call selx#SelXRefresh()
  else
    tab split
  endif
endfunction

" ----------------------------------
" The following mappings are used internally. If you change them then you
" will also ave to change the functions that invoke them; ie,
" the mapping (SelxAddN), (SelxAddNV) and (SelxClrN) and the functions
" selx#SelXGoToInit() and selx#SelXToggleSelInit()
"
" Note that although these mappings map the meta key, they will still work
" on installations that normally have problems with interpreting the meta
" key from the keyboard

" Set a specific highlight (rather than auto-selection)
nnoremap <silent> +<leader>& :call selx#SelXCmdExecute("call selx#SelXAdd('', selx#SelXCmdIndex)", '')<cr>
nnoremap <silent> +<leader>&<esc> <nul>

" Same thing for visual selection mode
vnoremap <silent> +<leader>& :call selx#SelXCmdExecute("call selx#VSetSearch('/', 1)", "call selx#SelXAdd(@/, selx#SelXCmdIndex)")<cr>
vnoremap <silent> +<leader>&<esc> <nul>

" Remove a specific highlight
nnoremap <silent> +<leader>: :call selx#SelXCmdExecute("call selx#SelXClr(selx#SelXCmdIndex)", '')<cr>
nnoremap <silent> +<leader>:<esc> <nul>

" Go to previous/next
nnoremap <silent> +<M-/> :call selx#SelXGoToNext(1, 0)<cr>
nnoremap <silent> +<M-/><esc> <nul>

" Toggle selection between 'whole word' and 'arbitrary' for specified highlight
nnoremap <silent> +<M-(> :call selx#SelXToggleSelNext()<cr>
nnoremap <silent> +<M-(><esc> <nul>

" ----------------------------------
" Refresh all selection highlights in current window - required when a new window is added
autocmd WinNew * :call selx#SelXRefresh()

" ------------------------------------------------------------------------------
" eof

