" plugin/selx.vim
"
" SelX - Richard Bentley-Green, 27/10/2020
"
" Additional highlighting functions - These work similarly to the normal '*'
" function but multiple highlights (in different colours) are supported.
" These highlights operate on a per-tab basis
"
if exists('g:Selx_loaded')
  finish
endif
let g:Selx_loaded = 1

" ----------------------------------
" Dummy version of the real function defined in autoload/selx.vim
" This is defined here to prevent autoload/selx.vim being brought-in
" prematurely (because it shall usually be called by the status line
" rendering on startup, even if SelX isnot active)
"
function selx#Statusline(bgcol)
  return ''
endfunction

" ----------------------------------
" Set highlight, and clear all highlights
nnoremap <silent> <plug>(SelxAdd) :call selx#SelXAdd('', 0)<cr>:call selx#SetShowSelX()<cr>
vnoremap <silent> <plug>(SelxAddV) :call selx#VSetSearch('/', 1)<cr>:call selx#SelXAdd(@/, 0)<cr>:call selx#SetShowSelX()<cr>
nnoremap <silent> <plug>(SelxClr) :call selx#SelXClr(0)<cr>:call selx#SetShowSelX()<cr>

" Set a specific highlight (rather than auto-selection)
nnoremap <silent> <plug>(SelxAddN) :call selx#SelXCmdInit("+\<leader>&")<cr>

" Same thing for visual selection mode
vnoremap <silent> <plug>(SelxAddNV) :call selx#SelXCmdInit("+\<Leader>&")<cr>gv

" Remove a specific highlight
nnoremap <silent> <plug>(SelxClrN) :call selx#SelXCmdInit("+\<Leader>:")<cr>

" Set a highlight
command -nargs=* SelX call selx#SelXAddCmd(<q-args>)
command -nargs=* SelXN call selx#SelXNAddCmd(<q-args>)

" Display details of active highlights
command SelXList call selx#ActiveList(get(g:, 'SelxCmdActiveList', 0))

" ----------------------------------
" Search for additional highlighted text

" Go to previous/next
nnoremap <silent> <plug>(SelxGotoFwd) :call selx#SelXGoToInit(1)<cr>
nnoremap <silent> <plug>(SelxGotoBkd) :call selx#SelXGoToInit(0)<cr>

" Repeat last search in same direction or (temporarily) in the reverse direction
nnoremap <silent> <plug>(SelxGotoNext) :call selx#SelXGoToNext(0, 0)<cr>
nnoremap <silent> <plug>(SelxGotoNextRev) :call selx#SelXGoToNext(0, 1)<cr>

" ----------------------------------
" Toggle selection between 'whole word' and 'arbitrary' for specified highlight
nnoremap <silent> <plug>(SelxToggleSelWhole) :call selx#SelXToggleSelInit()<cr>

" ----------------------------------
" Open a new tab with the current window full-screen within it. Also
" copy across any active 'SelX' highlights
nnoremap <silent> <plug>(SelxOpenInNewTab) :call selx#OpenInNewTab()<cr>

" ------------------------------------------------------------------------------
"  Optional modification to standard '*' and '#' operation

" Change the '*' behaviour so that it doesn't immediately jump to the next
" occurrence of the selected word (can still use 'n' and '#' to jump to
" next/previous)
nnoremap <silent> <plug>(SelxAddStd) :let @/= '\<'.expand('<cword>').'\>' <bar> set hls <cr>

" Make '*' and '#' work in visual mode. This makes the '*' operate as for normal
" mode; ie, it doesn't immediately jump to the next occurance
"xnoremap <silent> <plug>(SelxAddStdX) :call selx#VSetSearch('/', 1)<cr>/<C-R>=@/<cr><cr> :This version DOES jump to next occurance
xnoremap <silent> <plug>(SelxAddStdX) :call selx#VSetSearch('/', 1) <bar> set hls <cr>
xnoremap <silent> <plug>(SelxAddBkdStdX) :call selx#VSetSearch('?', 1)<cr>?<C-R>=@/<cr><cr>

" Go to next instance of word under cursor without changing search highlight
nnoremap <silent> <plug>(SelxGotoNextStd) :call selx#Search('', 1, 0)<cr>

" Same but for visual selection. This doesn't work very well because the target
" is not visually selected, making repeat searches rather awkward
vnoremap <silent> <plug>(SelxGotoNextStdV) :call selx#VSetSearch('/', 1)<cr>:call selx#Search(@/, 1, 0)<cr>

" Toggle selection between 'whole word' and 'arbitrary'
" text - Text to adjust
nnoremap <silent> <plug>(SelxToggleSelWholeStd) :let @/ = selx#ToggleSel(@/)<cr>

" ------------------------------------------------------------------------------
" Default mappings (for details of global options used here, see README)

if get(g:, 'SelxSetDefaultMaps', 1)
  nmap <silent> <leader>* <plug>(SelxAdd)
  vmap <silent> <leader>* <plug>(SelxAddV)
  nmap <silent> <leader>; <plug>(SelxClr)
  nmap <silent> <leader>& <plug>(SelxAddN)
  vmap <silent> <leader>& <plug>(SelxAddNV)
  nmap <silent> <leader>: <plug>(SelxClrN)

  if get(g:, 'SelxSearchMapingsUseMeta', has('nvim'))
    nmap <silent> <M-/> <plug>(SelxGotoFwd)
    nmap <silent> <M-?> <plug>(SelxGotoBkd)
    nmap <silent> <M-n> <plug>(SelxGotoNext)
    nmap <silent> <M-N> <plug>(SelxGotoNextRev)
  else
    nmap <silent> <leader>/ <plug>(SelxGotoFwd)
    nmap <silent> <leader>? <plug>(SelxGotoBkd)
    nmap <silent> <leader>n <plug>(SelxGotoNext)
    nmap <silent> <leader>N <plug>(SelxGotoNextRev)
  endif

  nmap <silent> <leader>( <plug>(SelxToggleSelWhole)
  nmap <silent> tt <plug>(SelxOpenInNewTab)

  if get(g:, 'SelxEnhancedStdHl', 1)
    nmap <silent> * <plug>(SelxAddStd)
    xmap <silent> * <plug>(SelxAddStdX)
    xmap <silent> # <plug>(SelxAddBkdStdX)
    nmap <silent> <leader><leader>n <plug>(SelxGotoNextStd)
    vmap <silent> <leader><leader>n <plug>(SelxGotoNextStdV)
    nmap <silent> <leader>) <plug>(SelxToggleSelWholeStd)
  endif
endif

" ------------------------------------------------------------------------------
" eof

